from os.path import (
    exists,
    join,
    dirname,
    basename,
    expanduser,
    isdir
)
import configparser, platform, json, os
from functools import lru_cache
from pathlib import Path

class Settings:
    # if path provided leads to a repo upstream
    is_repo      : bool = False
    is_module    : bool = False
    is_ymvas     : bool = False # if repos if fetched from ymvas
    is_main      : bool = False # if repo is main git@ymvas.com/vas/vas.git

    # paths
    root         : str  = None # /root             -> folder of the repo
    git          : str  = None # /root/.git        -> folder , if module -> /root/../../.git
    ymvas        : str  = None # /root/.ymvas      -> folder

    # repo info
    url          : str  = None # example  -> git@ymvas.com/vas/ymvas.git
    name         : str  = None # from url -> ymvas
    user         : str  = None # from url -> vas
    alias        : str  = None # from url -> vas/ymvas

    # ymvas important folders
    d_references : str = None
    d_commands   : str = None
    d_settings   : str = None
    d_tasks      : str = None
    d_schedules  : str = None
    d_secrets    : str = None
    d_finance    : str = None
    d_actions    : str = None
    d_endpoints  : str = None
    d_documents  : str = None
    d_contacts   : str = None
    d_spaces     : str = None

    # where are we going to complie the endpointss
    d_compile_prefix : str = None

    # submodules asociated
    modules: dict = {}

    # ymvas important files
    f_settings            : str = None
    f_settings_endpoints  : str = None
    f_settings_references : str = None
    f_settings_secrets    : str = None
    f_settings_tasks      : str = None
    f_settings_finance    : str = None
    f_settings_actions    : str = None

    d_default_compiles    : str = None

    def __init__( self , pwd ):
        root, is_repo = Settings._find_repo_root(pwd)

        self.root = root
        if is_repo:
            self.root = pwd
            self._setup_repo_paths()

        self.is_repo = is_repo

        if not exists( self.f_global_config ):
            os.makedirs(
                dirname( self.f_global_config ),
                exist_ok = True
            )

            with open(self.f_global_config,'w') as fs:
                from .__init__ import __version__
                data = {
                    "version" : __version__ ,
                }

                if self.is_main:
                    data['main']          = self.root
                    data['main-commands'] = self.d_commands

                fs.write(json.dumps(data, indent= 10))
        else:
            with open(self.f_global_config,'r') as fs:
                data = json.loads(fs.read())
                self.d_default_compiles = data.get('global-compile-dir',None)


    @property
    @lru_cache(maxsize=None)
    def ymvas_server_url(self):
        stg = self.get_global_settings()
        return str(stg.get(
            'ymvas-server-url',
            'ssh://git@'+self.ymvas_domain+'/{repo}.git'
        ))

    @property
    @lru_cache(maxsize=None)
    def ymvas_domain(self):
        stg = self.get_global_settings()
        return str(stg.get(
            'ymvas-domain',
            'ymvas.com'
        ))

    @staticmethod
    def _find_repo_root(pwd):
        _pwd_b = pwd

        ymf = join(pwd , '.ymvas' )
        ymp = join(pwd , '.ymvas' )
        git = join(pwd , '.git'   )

        _is_repo = True
        _search  = {}

        while not exists(ymf):
            _pwd_b = dirname(_pwd_b)

            _search[_pwd_b] = _search.get(_pwd_b,0) + 1
            ymf = join(_pwd_b,'.ymvas')

            if _search[_pwd_b] > 10 or exists(git):
                ymf = ymp
                _is_repo = False
                break

            git = join( _pwd_b , '.git'   )
            ymp = join( _pwd_b , '.ymvas' )

        return dirname(ymf), _is_repo


    def _setup_repo_paths( self ):
        _git      = join( self.root , '.git' )
        is_module = not isdir( _git )

        if not exists(_git):
            return

        if is_module:
            with open( _git ,'r' ) as f:
                _git = Path(f.read().split("gitdir:")[1].strip())
                _git = str( Path(self.root) / _git )

        self.git       = _git
        self.is_module = is_module


        # config
        cnf = configparser.ConfigParser()
        cnf.read( join(_git,'config') )

        _url = None
        for section in cnf.sections():
            if 'origin' in section and 'remote':
                _url = cnf[section].get('url',None)
                break


        user, name, is_ymvas = self._parse_url(_url)

        self.is_ymvas = is_ymvas
        self.user     = user
        self.name     = name
        self.url      = _url


        self.is_main = user == name
        self.alias   = f'{name}'
        if user is not None:
            self.alias = f'{user}/{name}'

        # /repo/.ymvas folders
        self.ymvas = join(self.root, '.ymvas' )

        self.d_references = join(self.ymvas, 'references' )
        self.d_commands   = join(self.ymvas, 'commands'   )
        self.d_settings   = join(self.ymvas, 'settings'   )
        self.d_tasks      = join(self.ymvas, 'tasks'      )
        self.d_schedules  = join(self.ymvas, 'schedules'  )
        self.d_secrets    = join(self.ymvas, 'secrets'    )
        self.d_finance    = join(self.ymvas, 'finance'    )
        self.d_actions    = join(self.ymvas, 'actions'    )
        self.d_endpoints  = join(self.ymvas, 'endpoints'  )

        # /repo folders if account
        if self.is_main:
            self.d_endpoints = join(self.root, 'endpoints' )
            self.d_documents = join(self.root, 'documents' )
            self.d_contacts  = join(self.root, 'contacts'  )
            self.d_finance   = join(self.root, 'finance'   )
            self.d_spaces    = join(self.root, 'spaces'    )

        # ymvas endpoints paths
        self.d_compile_prefix = self.alias
        if self.is_main:
            self.d_compile_prefix = user

        self.modules   = self.get_modules()

        # sttings files
        self.f_settings            = join(self.d_settings, 'settings.yaml'   )
        self.f_settings_endpoints  = join(self.d_settings, 'endpoints.yaml'  )
        self.f_settings_references = join(self.d_settings, 'references.yaml' )
        self.f_settings_secrets    = join(self.d_settings, 'secrets.yaml'    )
        self.f_settings_tasks      = join(self.d_settings, 'tasks.yaml'      )
        self.f_settings_finance    = join(self.d_settings, 'finance.yaml'    )
        self.f_settings_actions    = join(self.d_settings, 'actions.yaml'    )
        self.f_settings_schedules  = join(self.d_settings, 'schedules.yaml'  )



    def get_modules(self):
        file = join(self.root,'.gitmodules')

        if self.is_module:
            modules = join(".git","modules")
            fragment = modules + self.git.split(modules)[-1]
            main_path = self.git.replace(fragment,"")
            main_settings = Settings(main_path)
            return main_settings.get_modules()

        modules = {
            self.alias : {
                "root"   : True,
                "active" : True,
                "path"   : self.root ,
                "url"    : self.url  ,
                "name"   : self.name ,
                "user"   : self.user ,
            }
        }

        if not exists(file):
            return []

        cnf = configparser.ConfigParser()
        cnf.read( file )

        for s in cnf.sections():
            p = cnf[s].get('path'   , None )
            u = cnf[s].get('url'    , None )
            a = cnf[s].get('active' , 'True' ).lower().strip() == 'true'

            if not 'submodule' in s or p is None or u is None:
                continue

            user, name, is_ymvas = self._parse_url(u)
            if name is None:
                continue

            modules[f"{user}/{name}"] = {
                "root"   : False,
                "active" : a,
                "path"   : join( self.root , p ),
                "url"    : u     ,
                "name"   : name  ,
                "user"   : user  ,
            }

        return modules

    @property
    @lru_cache(maxsize=None)
    def f_global_config(self):
        system = platform.system()
        shell   = os.environ.get("SHELL", "")
        msystem = os.environ.get("MSYSTEM", "")

        # Detect Git Bash on Windows
        is_windows_git_bash = system == "Windows" and ("bash" in shell or msystem.startswith("MINGW"))
        if is_windows_git_bash:
            _dir = expanduser("~/.config")
        elif system == "Windows":
            _dir = os.environ.get("APPDATA", expanduser("~\\AppData\\Roaming"))
        elif system == "Darwin":  # macOS
            _dir = expanduser("~/Library/Application Support")
        else:  # Linux and other Unix-like systems
            _dir = os.environ.get("XDG_CONFIG_HOME", expanduser("~/.config"))

        return join(_dir, 'ymvas', 'config.json')

    @lru_cache(maxsize=None)
    def get_global_settings(self):
        if not exists(self.f_global_config):
            return {}
        try:
            with open(self.f_global_config,'r') as f:
                return json.loads(f.read())
        except Exception as e:
            print(e)
            return {}
        return {}

    def set_global_settings(self,stg):
        with open(self.f_global_config,'w') as fs:
            fs.write(json.dumps(stg))

    def _parse_url(self, url ):
        if url is None:
            return None, None, None

        is_ymvas = f"@{self.ymvas_domain}" in url
        name = basename(url).replace('.git','')

        # get repo user
        user1 = url.split('@')[-1]
        user  = user1.replace(basename(url),'')
        user  = basename(user.strip('/'))
        user  = user.split(':')[-1]
        user  = user.strip('/')
        user  = user if '/' not in user else None

        if user is not None:
            user  = None if user1.startswith( user ) else user

        return user, name, is_ymvas




import os ,shutil
from os.path import dirname, join, exists

_dir = dirname(__file__)

class Creator:

    def __init__(self, argv):
        self.src    = join(_dir,argv.init)
        self.target = argv.src

        if not exists(self.src):
            print(f'Invalid init [{argv.init}]')
            return

    def run(self):
        for r,_,files in os.walk(self.src):

            for f in files:
                src = join(r,f)
                dst = src.replace(self.src,self.target)

                if exists(dst):
                    continue

                dic = dirname(dst)
                os.makedirs(dic,exist_ok=True)

                shutil.copyfile(src, dst)



from pathlib import Path

from os.path import join, exists, dirname, isfile, isdir
import sys, os, yaml, json

from .references import Ref
from .settings   import Settings
from .compiler   import Compiler
from .templates.creator import Creator


class Operator:
    _refs  :dict = {}
    _spaces:dict = {}

    def __init__(self, pwd):
        self.settings = Settings(pwd)

    def command(self,args):

        commands_path = self.settings.d_commands
        if args.is_global or not self.settings.is_repo:
            commands_path = self.settings.get_global_settings().get(
                'main-commands',None
            )

            if commands_path is None:
                print("Global commands are not setup!")
                return

        # widnows git bash fixes
        commands_path = Path(commands_path)
        if str(commands_path).startswith("/c/") or \
           str(commands_path).startswith("\\c\\"):
            commands_path = Path("C:" + str(commands_path)[2:])

        if not commands_path.exists():
            print(f'Commands [.ymvas/commands/example.bash] is not setup for this repository!')
            return

        valid = [
            {"ext" : 'py'   , "run" : "python3" },
            {"ext" : 'bash' , "run" : "bash"    },
            {"ext" : 'sh'   , "run" : "sh"      },
        ]

        found = False
        for t in valid:
            ext  = t['ext']
            run  = t['run']

            file = join(commands_path,f"{args.command}.{ext}")

            if exists(file):
                found = True
                os.system(f'{run} {file}')

        if not found:
            raise Exception(f'Command [{args.command}] is not setup for this repository!')


    def clone(self,args):
        if args.repo is None:
            print('No repo specified!')
            return

        url = self.settings.ymvas_server_url.format(
            repo = args.repo
        )

        os.system(f"git clone {url}")

    def pull(self,argv):
        modules = self.settings.get_modules()
        modules = {k:v for k,v in modules.items() if not v['root'] and v['active']}

        os.system(
            f"git --git-dir={self.settings.git} "
            f"--work-tree={self.settings.root} fetch"
        )

        # for k,v in modules.items():
        #     p = v[ 'path' ]
        #     u = v[ 'url'  ]
        #
        #     if not exists(p):
        #         os.system(
        #             f"cd {self.settings.root} && "
        #             f"git --git-dir={self.settings.git} "
        #             f"--work-tree={self.settings.root} "
        #             f"submodule add {u} {p}"
        #         )


    def setup(self,argv):
        creator = Creator(argv)
        creator.run()

    def config(self,argv,args):
        if argv.action == 'set' and argv.is_global:
            stg = self.settings.get_global_settings()
            for a in args:
                if '=' not in a:
                    continue
                k = a.split('=')[0]
                v = a.replace(k+'=','').strip()

                if v == '':
                    continue
                k = k.strip('--')
                stg[k] = v
            self.settings.set_global_settings(stg)
            print(stg)
        elif argv.action == "show" and argv.is_global:
            print(self.settings.get_global_settings())

        elif argv.action == "get" and argv.is_global:
            for a in args:
                v = self.settings.get_global_settings().get(a,None)
                if v is None: continue
                print(v)

    def compile(self,args):
        if not exists(self.settings.d_endpoints):
            return

        self._append_refs(
            self.settings.alias,
            self.settings.d_references
        )

        self._spaces = self.settings.get_modules()

        Compiler(
            self ,
            join( args.compile_dir , self.settings.d_compile_prefix )
        ).run()

    def _append_refs(self, space, pref ):
        self._refs[space] = []

        if not exists(pref):
            return

        for r,_,files in os.walk( pref ):
            for f in files:
                self._refs[space].append(
                    Ref(join(r,f), self , space )
                )

        pass


    def refs(self,needle = None):
        if needle == None:
            for r in self._refs[self.alias]:
                yield r
            return

        for r in self._refs[self.alias]:
            fpath = r.fpath.replace(self.ref,'')
            if needle in fpath:
                yield r

    def __repr__(self):
        return "\n".join([
            f"config  : {self.settings.f_global_config}" ,
            "",
            f"[{self.settings.alias}]",
            f" - is-repo  : {self.settings.is_repo}",
            f" - is-ymvas : {self.settings.is_ymvas}",
            f" - is-main  : {self.settings.is_main}",
        ]) + "\n"



import yaml, json
from functools import cache
from .raw_tags import raw_tags

class Tag:
    _close_tag       = ''
    _link            = None
    _replace         = None

    def __init__(
        self              ,
        name       = ''   ,

        raw        = ''   ,
        contents   = None ,

        tabs       = ''   ,
        start      = 0    ,
        start_ends = None ,

        close      = None ,
        close_ends = None ,

        attrs      = {},
        slots      = [],
    ):

        self.name       = name
        self.tabs       = tabs

        self.start      = start
        self.start_ends = start_ends

        self.close      = close
        self.close_ends = close_ends

        self.raw        = raw
        self.contents   = contents

        self.attrs      = attrs
        self.slots      = slots

    def is_ref(self):
        return self.name == 'ref'
    def is_attr_true(self,attr):
        return self.attrs.get(attr,False) == True

    def outer_pos(self):
        if self.close_ends is None:
            return [self.start,self.start_ends]
        return [self.start,self.close_ends]

    def match(self,**kwargs):
        for k,v in kwargs.items():
            if getattr(self,k) != v:
                return False
        return True

    def get(self,attr,defaults_to=None):
        return self.attrs.get(attr,defaults_to)
    def __getitem__(self, item):
        return self.attrs.get(item,None)

    def get_paths(self):
        if self._link is not None:
            return self._link

        path = self.get('link')

        if path is None:
            return None
        path = path.split('/')
        self._link = {
            'name' : path[0],
            'path' : path[1:]
        }

        return self._link

    def ref(self):
        if self._link is None:
            self.get_paths()
        return self._link.get('name',None)

    def set_replace(self,v,lang=None):
        self._replace = v

        if lang == None:
            return

        try:
            if lang == 'yaml':
                v = yaml.safe_load( v )
            elif lang == 'json':
                v = json.load( v )
        except Exception:
            self._replace = v
            return

        path = self.get_paths()['path']

        for p in path:
            if str(p).isdigit() and isinstance(v,list):
                p = int(p)
            try:
                v = v[p]
            except Exception as e:
                px = self.get('link')
                v = f"{{key '{px}' is missing}}"
                break

        if isinstance(v, int):
            self._replace = str(v)
            return
        elif isinstance(v,str):
            self._replace = v
            return
        elif v is None:
            self._replace = ''
            return

        if self.is_attr_true('outer'):
            v = { path[-1] : v }

        if self.get('expose') != None \
           and self.get('expect') == None \
           and self.get('expose') in ['yaml','json']:
           lang = self.get('expose')

        if lang == 'yaml':
            v = yaml.dump(v,allow_unicode=True)
        elif lang == 'json':
            v = json.dumps(v)

        self._replace = v

    def __repr__(self):
        posi = [ self.start,self.start_ends,self.close,self.close_ends ]
        poss = ",".join([ str(x) for x in posi if x != None ])
        return f"<{self.name} {str(self.attrs)} pos='{poss}' />"

    def _render_counter(self):
        series = self.get( 'series' )
        if series is None:
            return str(self._replace)
        if str(series).isnumeric() and str(self._replace).isnumeric():
            s = int(series)
            n = int(self._replace)
            return str(int(n % s))
        return str(self._replace)

    def from_to(self,text,expect,expose):
        data = {}
        try:
            if expect == 'yaml':
                data = yaml.safe_load(text)
        except Exception as e:
            return text
        if expose == 'list':
            pass
        if expose == 'css':
            if not isinstance(data,dict):
                return text

            use_wpraps = [
                'px', '#', 'rgb', '-','.','rgb','var'
            ]

            def mk(dc,l=0):
                ls = []
                for k,v in dc.items():
                    if isinstance(v,dict):
                         v = mk(v,l+1)
                         ts = ('\t'*l)
                         ls.append( k + ("{\n" + ts ) + v + ts + '}' )
                         continue

                    _use = True
                    for u in use_wpraps:
                        if u in v:
                            _use = False
                            break
                    v = v.strip("\"'")

                    if _use:
                        v = f'"{v}"'

                    ls.append( k + ":" + v + ';' )
                return ('\n' + ('\t' * l)).join(ls)
            return mk(data)
        return text

    def render( self ):
        if not self.has_changes():
            return self.raw

        elif self.name == 'counter':
            return self._render_counter()

        if self['expose'] != None and self['expect'] != None:
            self._replace = self.from_to(
                self._replace, self['expect'], self['expose']
            )

        if self.is_ref():
            if self['ignore-spaces'] == None:
                self._replace = f"\n{self.tabs}".join(
                    str( self._replace ).split('\n')
                )


        if len(self.slots) == 0:
            return str(self._replace)

        s = Soup(str(self._replace))
        rslots = {}
        for t in s.tags:
            if t.name != 'slot': continue
            name = t.get('name','main')
            rslots[name] = rslots.get(name,[])
            rslots[name].append(t)

        for l in self.slots:
            name = l['attrs'].get('name','main')
            if name not in rslots:
                continue
            for r in rslots[name]:
                r._replace = l['contents']

        for t in s.tags:
            if t.name == 'slot' and t.has_changes():
                continue
            if t.name == 'slot':
                t._replace = ''

        return s.to_string()

    def has_changes(self):
        return self._replace != None

class Soup:
    def __init__(self, value:str):
        if value is None:
            self._val = ''
            self.tags = []
        self._val = value
        self.tags = [Tag(**t) for t in raw_tags(value)]

    def apply(self,args:dict,cbfn):
        for t in self.tags:
            good = True
            for k,v in args.items():
                a = getattr(t,k,None)
                if v != a:
                    good = False
                    break
            if good:
                cbfn(t)
        return self

    def refs(self):
        return [t for t in self.tags if t.is_ref()]

    def no_refs(self):
        for r in self.refs():
            r._replace = ""

    def find_by(self,**kwargs):
        arr = []

        attrs = {}
        if 'attrs' in kwargs:
            attrs = kwargs['attrs']
            del kwargs['attrs']

        has_attrs = len(list(attrs.keys())) != 0

        for t in self.tags:
            append = True

            for k,v in kwargs.items():
                a = getattr(t,k,None)
                if v != a:
                    append = False
                    break

            if has_attrs and append:
                links = t.attrs
                for a in attrs.keys():
                    if a not in links:
                        append = False
                        break
                    if attrs[a] != links[a]:
                        append = False
                        break

            if append:
                arr.append(t)

        return arr

    def to_string(self):
        nv,reach,_have_slots = '', 0, False

        # handle slots
        for t in self.tags:
            if len(t.slots) == 0:
                continue
            xt,et = t.outer_pos()
            for st in self.tags:
                if not st.has_changes():
                    continue

                _added = False
                for l in t.slots:
                    if st.raw in l['contents']:
                        l['contents'] = l['contents'].replace( st.raw, st._replace )
                        _added = True

                if _added:
                    st._replace = None

        for t in self.tags:
            s,e = t.outer_pos()
            if not t.has_changes():
                continue
            if s < reach:
                continue
            nv += self._val[reach:s]
            nv += t.render()
            reach = e

        nv += self._val[reach:len(self._val)]

        # print(nv)
        # print('#' * 80)
        return nv



import sys, os
from os.path import join, exists, dirname
from ymvas.operator import Operator

_defined_commands = ["info","clone","config","compile","init","pull"]
_defined_config   = ['set','get','show']
_defined_inits    = ['account']

red    = lambda x: f"\033[31m{x}\033[0m"
green  = lambda x: f"\033[32m{x}\033[0m"
yellow = lambda x: f"\033[33m{x}\033[0m"
blue   = lambda x: f"\033[34m{x}\033[0m"

def _keyval(args,short_key,key,default=None):
    trash = []
    value = default
    found = False
    for i, c in enumerate(args):
        if (c.startswith(key) or c.startswith(short_key)) and not found:
            value = c.replace(c.split('=')[0]+'=',"")
            found = True
        else:
            trash.append(c)
    return trash, value

def _flag(args,short_key,key,default=False):
    trash = []
    value = default
    found = False
    for i, c in enumerate(args):
        if (c == key or c == short_key) and not found:
            value = True
            found = True
        else:
            trash.append(c)
    return trash, value

def _default_info():
    print(f"{green('YMVAS')} - handle your life like a developer!")
    print(f"Usage : ")
    print(f"    {green('ym')} {{command}}   |  {yellow(','.join(_defined_commands))} ")
    print(f"    {green('ym')} --{{command}} |  {red('Is not allowed!')}      ")
    print( "")
    print( "File based commands placed like this:")
    print( "    ~/repo/.ymvas/commands/install.bash ")
    print( "    ~/repo/.ymvas/commands/install.py ")
    print( "    ~/repo/.ymvas/commands/install.sh ")
    print("")
    print(f"Can be easly executed like this: ")
    print(f"    {green('ym')} install        ")


def _compile_info():
    print(f"Usage : ")
    print(f"    {green('ym')} compile -c=\"~/destination/folder\"")
    print(f"    {green('ym')} compile -c=\"$(pwd)/test\"")
    print(f"    {green('ym')} compile --compile-dir=\"~/destination/folder\"")
    print(f"")
    print(yellow(f"You can also setup default compile directory with: "))
    print(yellow(f"    ym config set --global-compile-dir=\"~/your/path\""))


def _compile_no_repo():
    print(red(f"[{os.getcwd()}]: is not a valid repository!"))
    print("")


def _config_info():
    print(f"Usage : ")
    print(f"    {green('ym')} config     | {yellow(','.join(_defined_config))} ")
    print(f"    {green('ym')} config -g  | For global configuration ")
    print(f"")
    print(f"Example : ")
    print(f"    {green('ym')} config set -g  | For global configuration ")

def _init_info():
    print(f"Usage : ")
    print(f"    {green('ym')} init     | {yellow(','.join(_defined_inits))} ")
    print(f"")
    print(f"Example : ")
    print(f"    {green('ym')} init account ")


class Arguments:
    src         : str
    command     : str
    compile_dir : str
    is_global   : bool
    repo        : str
    action      : str

    init        : str

def run():
    args = sys.argv[1:]

    if len(args) == 0:
        return _default_info()

    argv = Arguments()
    argv.command = args[0]

    if argv.command == '--version' or argv.command == '-v':
        from .__init__ import __version__
        print(__version__)
        return

    if argv.command.startswith('-'):
        _default_info()
        return

    # extract src
    args , argv.src = _keyval(args[1:],'-s','--src',os.getcwd())
    cli = Operator(argv.src)

    if len(args) == 0 and argv.command == 'info':
        print(cli)
        return

    if argv.command == "clone":
        argv.repo = args[0] if '/' in args[0] else f"{args[0]}/{args[0]}"
        cli.clone(argv)
        return

    if argv.command == "init":
        if len(args) == 0:
            _init_info()
            return
        argv.init = args[0]
        cli.setup(argv)
        return

    args , argv.is_global = _flag(args,'-g','--global',False)

    if argv.command == 'config':
        if len(args) == 0 or args[0] not in _defined_config:
            _config_info()
            return

        argv.action = args[0]
        cli.config(argv,args[1:])
        return

    if argv.command not in _defined_commands:
        cli.command(argv)
        return

    if not cli.settings.is_repo:
        _compile_no_repo()
        print(cli)
        return

    if argv.command == 'compile':
        args , argv.compile_dir = _keyval( args,'-c','--compile-dir' )

        if cli.settings.d_default_compiles is None and argv.compile_dir is None:
            _compile_info()
            return

        cli.compile(argv)
        return

    if argv.command == 'pull':
        cli.pull(argv)
        return


if __name__ == '__main__':
    run()



from os.path import join, exists, dirname, basename, isdir
from pathlib import Path
import os, json, shutil, yaml

from .references import Ref

class Compiler:
    info  = []

    valid_formats = [
        "pdf",
        'png'
    ]

    def __init__(self, cli , out ):
        self.cli = cli
        stg = self.cli.settings

        # stored paths
        self._p_compiled_dir = out
        self._p_endpoints    = stg.d_endpoints
        self._p_repo         = stg.root
        self._p_repo_dot_git = stg.git

        # settings
        self._p_d_settings   = stg.d_settings
        self._is_account     = stg.is_main
        self._repo_alias     = stg.alias

        self.f_settings_endpoints = stg.f_settings_endpoints
        self.d_compile_prefix     = stg.d_compile_prefix

        # complier files
        self.cache_info_file = join(out,'config.json')
        self._cache          = {}
        self._has_cache      = False

        if exists(self.cache_info_file):
            try:
                f = open(self.cache_info_file,'r')
                self._cache = json.loads(f.read())
                f.close()
                self._has_cache = True
            except Exception:
                pass


    def _tracked_settings(self):
        tracked = " ".join([
            self._p_d_settings ,
        ])

        commited = os.popen(
            f"git --git-dir={self._p_repo_dot_git} "
            f"--work-tree={self._p_repo} diff "
            f"--name-only HEAD~1 HEAD "
            f"-- {tracked} "
        ).read().strip().split("\n")

        modified = os.popen(
            f"git --git-dir={self._p_repo_dot_git} "
            f"--work-tree={self._p_repo} status "
            f"-s {tracked} "
        ).read().strip().split("\n")

        files = []
        for f in modified:
            mtype = f[:3]
            file  = f[3:].strip()
            if '"' in f or "'" in f:
                continue
            if "R" in mtype:
                fxf  = file.split('->')
                file = fxf[-1].strip()
            files.append(file)

        return list(set(commited + files))

    def _tracked_files(self):
        if not self._has_cache:
            _files = []
            for r,_,files in os.walk( self._p_endpoints ):
                for f in files:
                    original = Path(join(r,f))
                    _files.append(
                        str(original.relative_to(Path(self._p_repo)))
                    )
            return _files


        tracked = " ".join([
            self._p_endpoints  ,
        ])

        commited = os.popen(
            f"git --git-dir={self._p_repo_dot_git} "
            f"--work-tree={self._p_repo} diff "
            f"--name-only HEAD~1 HEAD "
            f"-- {tracked} "
        ).read().strip().split("\n")

        modified = os.popen(
            f"git --git-dir={self._p_repo_dot_git} "
            f"--work-tree={self._p_repo} status "
            f"-s {tracked} "
        ).read().strip().split("\n")

        files = []
        for f in modified:
            mtype = f[:3]
            file  = f[3:].strip()

            if '"' in f or "'" in f:
                # skip strange files
                self.info.append({
                    "operation" : "file-skip-compile",
                    "file"      : file
                })
                continue

            if "R" in mtype:
                fxf  = file.split('->')
                file = fxf[-1].strip()

                self.info.append({
                    "operation" : "file-renamed",
                    "file"      : fxf[0].strip()
                })

            files.append(file)
        return list(set(commited + files))

    def _make_pdf(self, file, content, conf = {} ):
        try:
            import pdfkit

            file.write(pdfkit.from_string(
                content,
                False,
                options = {
                    'margin-top'    : str(conf.get( 'margin-top'    , '0.0in' )),
                    'margin-right'  : str(conf.get( 'margin-right'  , '0.0in' )),
                    'margin-bottom' : str(conf.get( 'margin-bottom' , '0.0in' )),
                    'margin-left'   : str(conf.get( 'margin-left'   , '0.0in' )),
                }
            ))

        except OSError as e:
            file.write(content.encode('utf-8'))
            print('[pdfkit] failed to create file, please install all the depedencies required!')

    def _make_svg_png(self,file,content,conf = {}):
        _content = content.encode('utf-8')
        try:
            from cairosvg import svg2png
            file.write(svg2png(bytestring=_content))
        except Exception as e:
            file.write(_content)
            print('[cairosvg] failed to create file, please install all the depedencies required!')


    def get_endpoints(self):
        if exists(self.f_settings_endpoints):
            with open(self.f_settings_endpoints,'r') as f:
                data = yaml.safe_load(f.read())
                if isinstance(data,dict):
                    return data
        return {}

    def run(self):
        os.makedirs( self._p_compiled_dir ,exist_ok =True)

        config = {
            "files" : {}
        }

        #
        files    = self._tracked_files()
        settings = self._tracked_settings()

        # settings modified in last commit
        _modified_settings = len(settings) != 0

        for f in files:
            ff = join(self._p_repo, f)
            if ff.startswith(self._p_d_settings):
                _modified_settings = True
                break

        endpoints = self.get_endpoints()

        # track modifications
        if _modified_settings:
            for mf, mf_info in self._cache.get('files',{}).items():

                # check current settings
                _curr_conf = endpoints.get(mf,{})
                _prev_conf = mf_info.get('conf',{})

                if json.dumps(_curr_conf) == json.dumps(_prev_conf):
                    # no changes
                    continue

                # add this file to be compiled again
                # because settings have changed
                origin = mf_info.get('origin',None)
                if origin != None:
                    files = list(set(files + [origin]))

                # remove previous compiled filefile
                _prev_compiled = mf_info.get('compiled',None)
                if _prev_compiled != None:
                    _prev_compiled = join(self._p_compiled_dir, _prev_compiled)
                    if exists(_prev_compiled):
                        os.remove(_prev_compiled)

                self.info.append({
                    "operation" : "file-removed",
                    "file"      : _prev_compiled
                })



        for f in files:
            ff     = join(self._p_repo, f)

            if isdir(ff):
                continue

            origin = str(Path(ff).relative_to(self._p_repo))
            name   = str(Path(ff).relative_to(self._p_endpoints))
            conf   = endpoints.get(name,{})

            format = conf.get('format',None)
            format = format if format in self.valid_formats else None

            e = Ref( ff, self.cli , self._repo_alias )

            nnamet = name
            if format != None:
                fname    = e.basename.split('.')[0] if '.' in e.basename else e.basename
                fname   += f'.{format}'
                nnamet  = name.strip(e.basename)
                nnamet += fname

            ntarget = join(self._p_compiled_dir, nnamet)
            vnam    = ntarget.replace(basename(ntarget),'')

            os.makedirs( vnam ,exist_ok = True )

            file = open(ntarget,'wb')
            if format == 'pdf':
                self._make_pdf(file,e.content,conf)
            elif format == 'png':
                self._make_svg_png(file,e.content,conf)
            else:
                file.write(e.content.encode('utf-8'))

            config['files'][name] = {
                "origin"   : origin,
                "compiled" : nnamet,
                "conf"     : conf
            }

            file.close()

        for f in self.info:
            # remove compiled if file was renamed
            if f.get('operation',None) != 'file-renamed':
                continue

            for mf, mf_info in self._cache.get('files',{}).items():
                origin = mf_info.get('origin',None)
                if origin != f:
                    continue

                _prev_compiled = mf_info.get('compiled',None)
                if _prev_compiled == None:
                    continue

                _prev_compiled = join(self._p_compiled_dir, _prev_compiled)
                if exists(_prev_compiled):
                    os.remove(_prev_compiled)
                    break


        if self._is_account and len(settings) != 0:
            posible_avatars = ["png","jpeg","jpg"]

            for a in posible_avatars:
                af = f"avatar.{a}"
                ap = join( '.ymvas', 'settings', af )

                config['avatar'] = {
                    "origin"  : ap ,
                    "display" : af ,
                }

                if ap in settings:
                    shutil.copy2(
                        join(self._p_repo , ap ),
                        join(self._p_compiled_dir, af )
                    )
                    break

        with open(self.cache_info_file,'w') as f:
            from .__init__ import __version__
            config['version'] = __version__

            f.write(json.dumps(config, indent = 10))

        domain = self.cli.settings.get_global_settings().get('domain',None)
        if domain is not None:
            for k,v in config["files"].items():
                c = v.get('compiled',None)
                if c is None: continue
                print( f"[COMPILER] : {domain}/{self.d_compile_prefix}/{c}" )




def tag_attrs(value):
    ref = value

    ref = ref.replace("/>","").replace('\n',' ').replace('>','')
    ref = " ".join([r for r in ref.split(' ') if r.strip() != '' and '<' not in r and '>' not in r])
    ref = ref.replace(" =","=").replace("= ","=")
    rei = [i for i, v in enumerate(ref) if v in ["'",'"']]

    dc = {"'":0,'"':0}
    lt = {"'":None,'"':None}
    pairs = []
    for c in rei:
        r = ref[c]
        n = dc[r]

        if n != 0: # close
            ref[c]
            dc[r] = 0
            o = "'" if r != "'" else '"'
            dc[o] = 0
            s,e = (lt[r],c)
            cnn = ref[s:e+1]
            pairs.append((s,e,cnn))
            lt[r] = None
        else:
            dc[r] += 1
            lt[r] = c

    dc = {str(i):v[2] for i,v in enumerate(pairs)}
    for k,v in dc.items():
        ref = ref.replace(v,f'{k}'.strip("'\"\n"))

    ref = ref.split(" ")
    data = {}
    for a in ref:
        g = a.split('=')
        data[g[0]] = True if len(g) < 2 else dc.get(g[1],None)
        i = data[g[0]]
        if isinstance(i,str):
            data[g[0]] = i[1:-1]

    return data

def raw_tags(value):
    slen = len(value)
    starters  = [(i,c) for i,c in enumerate(value) if c in ['<','>']]
    nstarters = len(starters)
    entries   = []

    for i, s in enumerate(starters):
        _next = slen if (i+1) == nstarters else starters[i+1]
        pos = s[0]
        typ = s[1]

        if typ == '<':
            name  = value[s[0]+1:_next[0]].split(' ')[0]
            name  = name.replace('\n','').replace('\t','')
            sname = name.replace("/",'')

            cnt = 0
            ends = None
            for j in range(i,nstarters):
                if starters[j][1] == "<":
                    cnt += 1
                if starters[j][1] == ">" and cnt != 0:
                    cnt -= 1
                if starters[j][1] == ">" and cnt == 0:
                    ends = starters[j][0]
                    break

            if ends == None:
                continue

            entries.append({
                "name" : sname,
                "has_slash" : sname != name,
                "start" : s[0],
                "ends" : ends+1,
                "ender" : False if len(name) == 0 else (name[0] == "/"),
                "stupid" : len(name) == 0,
                "simple" : False if len(name) == 0 else (name[-1] == "/"),
                "single" : value[s[0]:ends+1].endswith("/>")
            })




    tags = []
    for i,e in enumerate(entries):
        name,start = e['name'], e['start']
        start_ends = e['ends']

        if e['stupid'] or e['simple'] or e['single']:
            tags.append({
                "name" : name ,
                "tabs" : value[0:start].split('\n')[-1],
                "start" : start,
                "start_ends" : start_ends,
                "close" : None ,
                "close_ends" : None ,
            })
            continue
        if e['ender']:
            continue

        cnt,ends = 0,None
        for j in range(i,len(entries)):
            if entries[j]['name'] != name or entries[j]['single']:
                continue
            if not entries[j]['ender']:
                cnt += 1
            if entries[j]['ender'] and cnt != 0:
                cnt -= 1
            if entries[j]['ender'] and cnt == 0:
                ends = entries[j]
                break

        tags.append({
            "name" : name ,
            "tabs" : value[0:start].split('\n')[-1],
            "start" : start,
            "start_ends" : start_ends,
            "close" : None if ends is None else ends['start'],
            "close_ends" : None if ends is None else ends['ends'],
        })


    for t in tags:
        # t['start_tag'] = value[t['start']:t['start_ends']]
        # t['close_tag'] = value[t['close']:t['close_ends']]

        t[  'attrs'  ] = tag_attrs(
            value[t['start']:t['start_ends']]
        )

        t['raw'] = value[
            t['start']:t['start_ends']
        ] if (t['close_ends'] is None) else value[
            t['start']:t['close_ends']
        ]

        t['contents'] = None if (t['close_ends'] is None) else value[
            t['start_ends']:t['close']
        ]

    for t in tags:
        t['slots'] = []
        if t['contents'] == None:
            continue

        # get all sub tags that are slots
        for st in tags:
            if st['close_ends'] == None:
                continue
            if not (st['start'] > t['start'] and st['close_ends'] < t['close_ends']):
                continue
            if st['name'] == 'slot':
                t['slots'].append(st)

    return tags



import os, re
from os.path import join, basename, exists
from .soup import Soup

class Ref:
    cli       = None
    raw:str   = ''
    soup:Soup = None
    _content  = None

    _bloked   = False

    def __init__(self , fpath, cli ,space):
        self.space    = space
        self.cli      = cli
        self.fpath    = fpath
        self.basename = basename(fpath)

        parts         = self.basename.split(".")
        self.name     = parts[0]
        self.lang     = 'xml' if len(parts) == 1 else parts[-1]

    def __repr__(self):
        return f"<ref link='{self.name}' />"


    def _compile(self):
        if self._content != None:
            return

        with open(self.fpath,'r') as f:
            self.raw = f.read()

        self.soup = Soup(self.raw)
        self._bloked = True
        for t in self.soup.refs():
            s    = t.get('space',self.space)
            name = t.ref()

            # print(s,name,t)
            if s in self.cli._spaces \
               and not s in self.cli._refs:

                # append subrepos references
                self.cli._append_refs( s,
                    join( self.cli._spaces[s]['path'], '.ymvas', 'references' )
                )

            for r in self.cli._refs.get(s,[]):
                if r.name == name:
                    if r is self:
                        t.set_replace(r.raw,r.lang)
                        break
                    t.set_replace(r.content,r.lang)
                    break

        self._content = self.soup.to_string()
        self._bloked  = False

    @property
    def content(self):
        if self._bloked:
            soup = Soup(self.raw)
            for t in soup.refs():
                t.set_replace('')
            return soup.to_string()
        self._compile()
        return self._content
